<?php
/**
* This is a class for computing the difference between two files, two
* strings, or any other two lists of things.  It uses an  intelligent
* algorithm similar to (or identical to) the one used by the Unix `diff'
* program. It is designed to find the *smallest possible* set of
* differences.  
* 
* 
* This class is based on the Perl Module Algorithm::Diff, which was 
* written by Ned Konz and Mark-Jason Dominus. Also the documentation of 
* this module was written by them.
* Thank you two :)
* 
*/
class diff
{
	/**
	* stores the computed diff
	* @access private
	*/
	var $retval=array();
	
	/**
	* do we want to ignore blanks/tabs ?
	* @access private
	*/
	var $ignoreblanks=0;
	
	/**
	* Compares two arrays $a and $b ($a and $b could be the output of file()).
	* It computes the smallest set of additions and deletions necessary to turn
	* $a to $b.  
	*
	* @param array a
	* @param array b
	*/
	function diff($a, $b, $ignoreblanks=0)
	{
		$retval=array();
		$hunk=array();
		$this->ignoreblanks=intval($ignoreblanks);
		$this->traverse_sequences($a, $b, $retval, $hunk);
		$this->matchCallback($retval,$hunk);
		
		$this->retval=$retval;
		unset($retval);
	}	
	
	
	/**
	* Returns the description of the computed changes.
	* The returned value is a multidimensional array.
	* each element of the array is a hunk of differences.
	* each element of a hunk contains information about 
	* one line that should be added, removed or replaced.
	* 
	* Example: the two arrays you want to compare are the
	* following:
	* a b c e h j l m n p
	* b c d e f j k l m r s t
	* 
	* The returned list would look like this:
	*
	* [ 
	*   [ [ '-', 0 ] ],       
	*   [ [ '+', 2 ] ],
	*   [ [ '-', 4 ] , 
	*     [ '+', 4 ] ],
	*   [ [ '+', 6 ] ],
	*   [ [ '-', 8 ], 
	*     [ '-', 9 ], 
	*     [ '+', 9 ], 
	*     [ '+', 10 ], 
	*     [ '+', 11 ],
	*   ]
	* ]
	*
	* There are five hunks here. The first hunk says that 
	* the a at position 0 of the first array should be deleted (-).
	* The second hunk says that the d at position 2 of the second 
	* array should be inserted (+). The third hunk says that 
	* the h at position 4 of the first array should be removed 
	* and replaced with the f from position 4 of the second array. 
	* The other two hunks similarly.
	*
	* 
	* @return array diff
	*/
	function getdiff()
	{
		return $this->retval;	
	}
	
	
	/**
	* Create an associative array that maps each element of $aCollection to 
	* the set of positions it occupies in $aCollection, restricted to 
	* the elements within the range of indexes specified by $start and $end.
	*
	* my $hashRef = _withPositionsOfInInterval( \@array, $start, $end, $keyGen );
	*
	* @param array
	* @param integer start
	* @param integer end
	* @return array
	*/
	function _withPositionsOfInInterval($aCollection, $start, $end)
	{
		$d=array();
		$index;
		for($index=$start; $index<=$end; $index++)
		{
			$element = $aCollection[$index];
			$key = $this->keyGen($element);
			if(isset($d[$key]))
			{
				array_unshift($d[$key], $index);
			}
			else
			{
				$d[$key] = array($index);	
			}
			
		}
		return $d;
	}
	


	/**
	* Find the place at which aValue would normally be inserted into the array. If
	* that place is already occupied by aValue, do nothing, and return undef. If
	* the place does not exist (i.e., it is off the end of the array), add it to
	* the end, otherwise replace the element at that point with aValue.
	* It is assumed that the array's values are numeric.
	* This is where the bulk (75%) of the time is spent in this module, so try to
	* make it fast
	* 
	* count() on large arrays seems to be a little bit slow, so I used
	* a $arraycount. The use of a flipped array (see array_flip()) eliminates
	* the need to call array_search() which is very slow.
	* $arraycount and $fliparray will be updated everytime the array is updated.
	* 
	* @param array
	* @param array flipArray
	* @param integer aValue
	* @param integer high
	* @return mixed
	*/
	function _replaceNextLargerWith(&$array, &$fliparray, &$arraycount, $aValue, $high)
	{
		// using a $fliparray (which is nothing but array_flip($array)
		// eliminates the need to call array_search() which is very slow :(
		
		// so $aValue is already in $array, return false ..
		if(isset($fliparray[$aValue]) || $fliparray[$aValue])
		{
			return false;
		}
		
		// we need to add aValue
		else
		{
			$lastIndex=$arraycount-1;
			$high=($high) ? ($high) : ($lastIndex);

			// add aValue at the end of the array
			if($high==-1 || $aValue>$array[$lastIndex])
			{
				$array[]=$aValue;
				$fliparray[$aValue] = $lastIndex+1;
				$arraycount++;
				return $lastIndex+1;
			}
			// we must find the insertion point first ...
			else
			{
				// binary search for insertion point...
				$low = 0;
				$index;
				$found;
				while($low <= $high)
				{
					$index = (int)(( $high + $low ) / 2);
					$found = $array[$index];
					if($aValue==$found)
					{
						return false;
					}
					elseif($aValue > $found)
					{
						$low = $index + 1;
					}
					else
					{
						$high = $index - 1;
					}
				}
				// now insertion point is in $low.
				
				// since the next larger value will
				// be overwritten, we must update the
				// flipped array (or use array_flip)
				if(isset($fliparray[($array[$low])])) unset($fliparray[($array[$low])]);
				else $arraycount++;
				$fliparray[$aValue] = $low;
				
				// overwrite next larger value
				$array[$low] = $aValue;

				return $low;
			}
		}
	}
	
	
	



	
	
	/**
	* This method computes the longest common subsequence in $a and $b.
	* 
	* Result is array, whose contents is such that
	* $a[ $i ] == $b[ $result[ $i ] ]
	* foreach $i in ( 0 .. count($result)-1 ) if $result[ $i ] is defined.
	* 
	* By default, comparisons will use strcmp() and elements will be turned into
	* keys using the md5 hash.
	* 
	* Additional parameters, if any, will be passed to the key generation routine.
	*/
	function _longestCommonSubsequence(&$a, &$b)
	{
		$aStart=0;
		end($a);
		$aFinish=key($a);
		reset($a);
		
		$bStart=0;
		end($b);
		$bFinish=key($b);
		reset($b);
		
		$matchVector=array();
		
		// First we prune off any common elements at the beginning
		while ( $aStart <= $aFinish && $bStart <= $bFinish && $this->compare($a[$aStart],$b[$bStart]))
		{
			$matchVector[$aStart++] = $bStart++;
		}
	
		// now the end
		while ($aStart <= $aFinish && $bStart <= $bFinish && $this->compare($a[$aFinish], $b[$bFinish]))
		{
			$matchVector[$aFinish--] = $bFinish--;
		}
		
		// Now compute the equivalence classes of positions of elements
		$bMatches = $this->_withPositionsOfInInterval($b, $bStart, $bFinish);
		$thresh = array();
		$threshflip = array();
		$threshcount = 0;
		$links  = array();
		
		$i;
		$ai;
		$j;
		$k;
		for($i=$aStart; $i<=$aFinish; $i++)
		{
			$ai = $this->keyGen($a[$i]);
			if(isset($bMatches[$ai]))
			{
				$k = 0;
				foreach($bMatches[$ai] as $j)
				{
					// optimization: most of the time this will be true
					if ($k && $thresh[$k] > $j && $thresh[$k-1] < $j)
					{
						// may be $thresh[$k] is already defined.
						// if so, we must delete the old value
						// from the flipped array
						if(isset($threshflip[$thresh[$k]])) unset($threshflip[$thresh[$k]]);
						else $threshcount++;
						$thresh[$k] = $j;
						$threshflip[$j] = $k;
					}
					else
					{
						$k = $this->_replaceNextLargerWith($thresh, $threshflip, $threshcount, $j, $k);
					}
					
					$links[$k] = array((($k) ? ($links[$k-1]) : (null)), $i, $j);
					$loops++;
				}
			}
		}
		unset($threshflip);
		
		if($thresh)
		{
			end($thresh);
			$lasttresh=key($thresh);
			reset($thresh);
			
			for($link = $links[$lasttresh]; $link; $link=$link[0])
			{
				$matchVector[$link[1]] = $link[2];
			}
		}
		return $matchVector;
	}








	/**
	* traverse_sequences is the most general facility provided by this module.
	* Imagine that there are two arrows. Arrow A points to an element of array A, 
	* and arrow B points to an element of the arry B. Initially, the arrows point 
	* to the first elements of the respective arrays. traverse_sequences will 
	* advance the arrows through the arrayes one element at a time, calling 
	* an appropriate callback function before each advance. It willadvance the 
	* arrows in such a way that if there are equal elements $A[$i] and $B[$j] 
	* which are equal and which are part of the LCS, there will be some moment 
	* during the execution of traverse_sequences when arrow A is pointing to $A[$i] 
	* and arrow B is pointing to $B[$j]. When this happens, traverse_sequences 
	* will call the MATCH callback function and then it will advance both arrows. 
	* 
	* Otherwise, one of the arrows is pointing to an element of its array 
	* that is not part of the LCS. traverse_sequences will advance that arrow 
	* and will call the DISCARD_A or the DISCARD_B callback, depending on which 
	* arrow it advanced. If both arrows point to elements that are not part of 
	* the LCS, then traverse_sequences will advance one of them and call the 
	* appropriate callback, but it is not specified which it will call. 
	* 
	* 
	* @param array a
	* @param array b
	* @param array retval
	* @param arrray hunk
	* @return void
	*/
	function traverse_sequences(&$a, &$b, &$retval, &$hunk)
	{
		$callbacks = (($callbacks) ? ($callbacks) : (array()));
		
		$matchVector = $this->_longestCommonSubsequence($a, $b);
		ksort($matchVector);
		end($matchVector);
		$endmatchVector=key($matchVector);
		reset($matchVector);

		// Process all the lines in $matchVector
		end($a);
		$lastA=key($a);
		reset($a);
		end($b);
		$lastB=key($b);
		reset($b);
		$bi=0;
		$ai;
		
		
		for($ai=0; $ai<=$endmatchVector; $ai++)
		{
			$bLine = $matchVector[$ai];
			if(isset($matchVector[$ai])) // matched
			{
				while($bi < $bLine)
				{
						$this->discardBCallback($hunk, $b, $bi);
						$bi++;
				}
				$this->matchCallback($retval, $hunk);
				$bi++;
			}
			else
			{
				$this->discardACallback($hunk, $a, $ai);
			}
		}
		
		// The last entry (if any) processed was a match.
		// $ai and $bi point just past the last matching lines in their sequences.
		while($ai <= $lastA || $bi <= $lastB)
		{
			// last A?
			if($ai == $lastA + 1 && $bi <= $lastB)
			{
				while($bi <= $lastB)
				{
					$this->discardBCallback($hunk, $b, $bi);
					$bi++;
				}
			}
	
			// last B?
			if($bi == $lastB + 1 && $ai <= $lastA)
			{
				while($ai <= $lastA)
				{
					$this->discardACallback($hunk, $a, $ai);
					$ai++;
				}
			}
	
			if($ai <= $lastA)
			{
				$this->discardACallback($hunk, $a, $ai);
				$ai++;
			}
			if($bi <= $lastB)
			{
				$this->discardBCallback($hunk, $b, $bi);
				$bi++;			
			}
		}
	}




	/**
	* keyGen function which generates the md5 hash
	* of a value so it can be used as a key for an
	* associative array.
	* 
	* @param string a
	* @return string md5-hash
	*/
	function keyGen($a)
	{
		return md5($this->ignoreblanks ? trim($a) : $a);
		//return md5(trim($a));
	}
	
	/**
	* compares two strings $a and $b
	* return true if they equal, if not false.
	* 
	* @param string a
	* @param string b
	* @return boolean equal
	*/
	function compare($a,$b)
	{
		return ((strcmp(($this->ignoreblanks ? trim($a) : $a),($this->ignoreblanks ? trim($b) : $b))==0) ? (true) : (false));
		//return ((strcmp(trim($a),trim($b))==0) ? (true) : (false));
	}
	
	
	/**
	* --
	*/
	function matchCallback(&$retval, &$hunk)
	{
		if(count($hunk)>0)
		{
			$retval[]=$hunk;
		}
		$hunk=array();
	}
	/**
	* --
	*/
	function discardACallback(&$hunk, &$a, $x)
	{
		$hunk[]=array('sign' => '-', 'item_no' => $x);
	}
	/**
	* --
	*/
	function discardBCallback(&$hunk, &$b, $x)
	{
		$hunk[]=array('sign' => '+', 'item_no' => $x);
	}
}
?>